@@ -1,5 +1,5 @@
 /*
- *  $Id: main_macosx.mm,v 1.16 2006/05/08 16:56:07 gbeauche Exp $
+ *  $Id: main_macosx.mm,v 1.17 2007/01/13 18:21:30 gbeauche Exp $
  *
  *  main_macosx.mm -	Startup code for MacOS X
  *						Based (in a small way) on the default main.m,
@@ -78,7 +78,6 @@ int CPUType;
 bool CPUIs68060;
 int FPUType;
 bool TwentyFourBitAddressing;
-bool ThirtyThreeBitAddressing = false;
 
 
 // Global variables
@@ -114,31 +113,15 @@ static bool lm_area_mapped = false;	// Flag: Low Memory area mmap()ped
  *  Helpers to map memory that can be accessed from the Mac side
  */
 
-// NOTE: VM_MAP_33BIT is only used when compiling a 64-bit JIT on specific platforms
+// NOTE: VM_MAP_32BIT is only used when compiling a 64-bit JIT on specific platforms
 void *vm_acquire_mac(size_t size)
 {
-	void *m = vm_acquire(size, VM_MAP_DEFAULT | VM_MAP_33BIT);
-#ifdef USE_33BIT_ADDRESSING
-	if (m == VM_MAP_FAILED) {
-		printf("WARNING: Cannot acquire memory in 33-bit address space (%s)\n", strerror(errno));
-		ThirtyThreeBitAddressing = false;
-		m = vm_acquire(size);
-	}
-#endif
-	return m;
+	return vm_acquire(size, VM_MAP_DEFAULT | VM_MAP_32BIT);
 }
 
 static int vm_acquire_mac_fixed(void *addr, size_t size)
 {
-	int ret = vm_acquire_fixed(addr, size, VM_MAP_DEFAULT | VM_MAP_33BIT);
-#ifdef USE_33BIT_ADDRESSING
-	if (ret < 0) {
-		printf("WARNING: Cannot acquire fixed memory in 33-bit address space (%s)\n", strerror(errno));
-		ThirtyThreeBitAddressing = false;
-		ret = vm_acquire_fixed(addr, size);
-	}
-#endif
-	return ret;
+	return vm_acquire_fixed(addr, size, VM_MAP_DEFAULT | VM_MAP_32BIT);
 }
 
 
@@ -303,11 +286,6 @@ bool InitEmulator (void)
 	// Initialize VM system
 	vm_init();
 
-#ifdef USE_33BIT_ADDRESSING
-	// Speculatively enables 33-bit addressing
-	ThirtyThreeBitAddressing = true;
-#endif
-
 #if REAL_ADDRESSING
 	// Flag: RAM and ROM are contigously allocated from address 0
 	bool memory_mapped_from_zero = false;
