@@ -31,12 +31,15 @@
 #endif
 
 // Glue for SDL and X11 support
+#ifdef TEST_VOSF_PERFORMANCE
+#define MONITOR_INIT			/* nothing */
+#else
 #ifdef USE_SDL_VIDEO
 #define MONITOR_INIT			SDL_monitor_desc &monitor
 #define VIDEO_DRV_WIN_INIT		driver_window *drv
 #define VIDEO_DRV_DGA_INIT		driver_fullscreen *drv
-#define VIDEO_DRV_LOCK_PIXELS	if (SDL_MUSTLOCK(drv->s)) SDL_LockSurface(drv->s)
-#define VIDEO_DRV_UNLOCK_PIXELS	if (SDL_MUSTLOCK(drv->s)) SDL_UnlockSurface(drv->s)
+#define VIDEO_DRV_LOCK_PIXELS	SDL_VIDEO_LOCK_SURFACE(drv->s)
+#define VIDEO_DRV_UNLOCK_PIXELS	SDL_VIDEO_UNLOCK_SURFACE(drv->s)
 #define VIDEO_DRV_DEPTH			drv->s->format->BitsPerPixel
 #define VIDEO_DRV_WIDTH			drv->s->w
 #define VIDEO_DRV_HEIGHT		drv->s->h
@@ -66,6 +69,11 @@
 #define VIDEO_DRV_HEIGHT		VIDEO_DRV_IMAGE->height
 #define VIDEO_DRV_ROW_BYTES		VIDEO_DRV_IMAGE->bytes_per_line
 #endif
+#endif
+
+// Prototypes
+static void vosf_do_set_dirty_area(uintptr first, uintptr last);
+static void vosf_set_dirty_area(int x, int y, int w, int h, int screen_width, int screen_height, int bytes_per_row);
 
 // Variables for Video on SEGV support
 static uint8 *the_host_buffer;	// Host frame buffer in VOSF mode
@@ -217,11 +225,20 @@ static bool video_vosf_profitable(void)
 	int64 durations[VOSF_PROFITABLE_TRIES];
 	int mean_duration = 0;
 
+#ifdef SHEEPSHAVER
+	const bool accel = PrefsFindBool("gfxaccel");
+#else
+	const bool accel = false;
+#endif
+
 	for (int i = 0; i < VOSF_PROFITABLE_TRIES; i++) {
 		uint64 start = GetTicks_usec();
 		for (int p = 0; p < mainBuffer.pageCount; p++) {
 			uint8 *addr = (uint8 *)(mainBuffer.memStart + (p * mainBuffer.pageSize));
-			addr[0] = 0; // Trigger Screen_fault_handler()
+			if (accel)
+				vosf_do_set_dirty_area((uintptr)addr, (uintptr)addr + mainBuffer.pageSize - 1);
+			else
+				addr[0] = 0; // Trigger Screen_fault_handler()
 		}
 		int64 duration = GetTicks_usec() - start;
 		mean_duration += duration;
@@ -321,6 +338,73 @@ static void video_vosf_exit(void)
 }
 
 
+/*
+ * Update VOSF state with specified dirty area
+ */
+
+static void vosf_do_set_dirty_area(uintptr first, uintptr last)
+{
+	const int first_page = (first - mainBuffer.memStart) >> mainBuffer.pageBits;
+	const int last_page = (last - mainBuffer.memStart) >> mainBuffer.pageBits;
+	uint8 *addr = (uint8 *)(first & -mainBuffer.pageSize);
+	for (int i = first_page; i <= last_page; i++) {
+		if (PFLAG_ISCLEAR(i)) {
+			PFLAG_SET(i);
+			vm_protect(addr, mainBuffer.pageSize, VM_PAGE_READ | VM_PAGE_WRITE);
+		}
+		addr += mainBuffer.pageSize;
+	}
+}
+
+static void vosf_set_dirty_area(int x, int y, int w, int h, int screen_width, int screen_height, int bytes_per_row)
+{
+	if (x < 0) {
+		w -= -x;
+		x = 0;
+	}
+	if (y < 0) {
+		h -= -y;
+		y = 0;
+	}
+	if (w <= 0 || h <= 0)
+		return;
+	if (x + w > screen_width)
+		w -= (x + w) - screen_width;
+	if (y + h > screen_height)
+		h -= (y + h) - screen_height;
+	LOCK_VOSF;
+	if (bytes_per_row >= screen_width) {
+		const int bytes_per_pixel = bytes_per_row / screen_width;
+		if (bytes_per_row <= mainBuffer.pageSize) {
+			const uintptr a0 = mainBuffer.memStart + y * bytes_per_row + x * bytes_per_pixel;
+			const uintptr a1 = mainBuffer.memStart + (y + h - 1) * bytes_per_row + (x + w - 1) * bytes_per_pixel;
+			vosf_do_set_dirty_area(a0, a1);
+		} else {
+			for (int j = y; j < y + h; j++) {
+				const uintptr a0 = mainBuffer.memStart + j * bytes_per_row + x * bytes_per_pixel;
+				const uintptr a1 = a0 + (w - 1) * bytes_per_pixel;
+				vosf_do_set_dirty_area(a0, a1);
+			}
+		}
+	} else {
+		const int pixels_per_byte = screen_width / bytes_per_row;
+		if (bytes_per_row <= mainBuffer.pageSize) {
+			const uintptr a0 = mainBuffer.memStart + y * bytes_per_row + x / pixels_per_byte;
+			const uintptr a1 = mainBuffer.memStart + (y + h - 1) * bytes_per_row + (x + w - 1) / pixels_per_byte;
+			vosf_do_set_dirty_area(a0, a1);
+		} else {
+			for (int j = y; j < y + h; j++) {
+				const uintptr a0 = mainBuffer.memStart + j * bytes_per_row + x / pixels_per_byte;
+				const uintptr a1 = mainBuffer.memStart + j * bytes_per_row + (x + w - 1) / pixels_per_byte;
+				vosf_do_set_dirty_area(a0, a1);
+			}
+		}
+	}
+	mainBuffer.dirty = true;
+	UNLOCK_VOSF;
+}
+
+
 /*
  * Screen fault handler
  */
@@ -336,8 +420,10 @@ bool Screen_fault_handler(sigsegv_address_t fault_address, sigsegv_address_t fau
 	if (((uintptr)addr - mainBuffer.memStart) < mainBuffer.memLength) {
 		const int page  = ((uintptr)addr - mainBuffer.memStart) >> mainBuffer.pageBits;
 		LOCK_VOSF;
-		PFLAG_SET(page);
-		vm_protect((char *)(addr & -mainBuffer.pageSize), mainBuffer.pageSize, VM_PAGE_READ | VM_PAGE_WRITE);
+		if (PFLAG_ISCLEAR(page)) {
+			PFLAG_SET(page);
+			vm_protect((char *)(addr & -mainBuffer.pageSize), mainBuffer.pageSize, VM_PAGE_READ | VM_PAGE_WRITE);
+		}
 		mainBuffer.dirty = true;
 		UNLOCK_VOSF;
 		return true;
@@ -385,6 +471,7 @@ There are two cases to check:
 	than pageCount.
 */
 
+#ifndef TEST_VOSF_PERFORMANCE
 static void update_display_window_vosf(VIDEO_DRV_WIN_INIT)
 {
 	VIDEO_MODE_INIT;
@@ -431,6 +518,7 @@ static void update_display_window_vosf(VIDEO_DRV_WIN_INIT)
 	}
 	mainBuffer.dirty = false;
 }
+#endif
 
 
 /*
@@ -438,6 +526,7 @@ static void update_display_window_vosf(VIDEO_DRV_WIN_INIT)
  *	(only in Real or Direct Addressing mode)
  */
 
+#ifndef TEST_VOSF_PERFORMANCE
 #if REAL_ADDRESSING || DIRECT_ADDRESSING
 static void update_display_dga_vosf(VIDEO_DRV_DGA_INIT)
 {
@@ -573,6 +662,7 @@ static void update_display_dga_vosf(VIDEO_DRV_DGA_INIT)
 	mainBuffer.dirty = false;
 }
 #endif
+#endif
 
 #endif /* ENABLE_VOSF */
 
