@@ -15,16 +15,16 @@ extern "C" {
 
 int slirp_init(void);
 
-void slirp_select_fill(int *pnfds, 
-                       fd_set *readfds, fd_set *writefds, fd_set *xfds);
+int slirp_select_fill(int *pnfds, 
+					  fd_set *readfds, fd_set *writefds, fd_set *xfds);
 
 void slirp_select_poll(fd_set *readfds, fd_set *writefds, fd_set *xfds);
 
-void slirp_input(const uint8_t *pkt, int pkt_len);
+void slirp_input(const uint8 *pkt, int pkt_len);
 
 /* you must provide the following functions: */
 int slirp_can_output(void);
-void slirp_output(const uint8_t *pkt, int pkt_len);
+void slirp_output(const uint8 *pkt, int pkt_len);
 
 int slirp_redir(int is_udp, int host_port, 
                 struct in_addr guest_addr, int guest_port);
