@@ -6,6 +6,7 @@
  */
 
 #define WANT_SYS_IOCTL_H
+#include <stdlib.h>
 #include <slirp.h>
 
 u_int curtime, time_fasttimo, last_slowtimo, detach_time;
@@ -137,7 +138,7 @@ struct quehead {
 	struct quehead *qh_rlink;
 };
 
-inline void
+void
 insque(a, b)
 	void *a, *b;
 {
@@ -150,7 +151,7 @@ insque(a, b)
 	= (struct quehead *)element;
 }
 
-inline void
+void
 remque(a)
      void *a;
 {
@@ -821,8 +822,8 @@ void
 fd_nonblock(fd)
 	int fd;
 {
-#ifdef FIONBIO
-	int opt = 1;
+#if defined USE_FIONBIO && defined FIONBIO
+	ioctlsockopt_t opt = 1;
 	
 	ioctlsocket(fd, FIONBIO, &opt);
 #else
@@ -838,8 +839,8 @@ void
 fd_block(fd)
 	int fd;
 {
-#ifdef FIONBIO
-	int opt = 0;
+#if defined USE_FIONBIO && defined FIONBIO
+	ioctlsockopt_t opt = 0;
 	
 	ioctlsocket(fd, FIONBIO, &opt);
 #else
