@@ -11,10 +11,10 @@
  *
  *  Copyright 1999, 2000, 2001, 2002, 2003 Ian Piumarta
  *
- *  Adaptations and enhancements for AMD64 support, Copyright 2003
+ *  Adaptations and enhancements for AMD64 support, Copyright 2003-2006
  *    Gwenole Beauchesne
  *
- *  Basilisk II (C) 1997-2005 Christian Bauer
+ *  Basilisk II (C) 1997-2006 Christian Bauer
  *  
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -191,7 +191,7 @@ enum {
  */
 
 #define _r0P(R)		((int)(R) == (int)X86_NOREG)
-#define _rIP(R)		((int)(R) == (int)X86_RIP)
+#define _rIP(R)		(X86_TARGET_64BIT ? ((int)(R) == (int)X86_RIP) : 0)
 
 #if X86_FLAT_REGISTERS
 #define _rC(R)		((R) & 0xf0)
@@ -201,7 +201,7 @@ enum {
 #else
 #define _rN(R)		((R) & 0x07)
 #define _rR(R)		(int(R))
-#define _rXP(R)		(int(R) > 7)
+#define _rXP(R)		(_rR(R) > 7 && _rR(R) < 16)
 #endif
 
 #if !defined(_ASM_SAFETY) || ! X86_FLAT_REGISTERS
@@ -248,10 +248,10 @@ typedef unsigned int	_ul;
 #define _SL(X)		((_sl  )(unsigned long)(X))
 #define _UL(X)		((_ul  )(unsigned long)(X))
 
-# define _PUC(X)	((_uc *)(X))
-# define _PUS(X)	((_us *)(X))
-# define _PSL(X)	((_sl *)(X))
-# define _PUL(X)	((_ul *)(X))
+#define _PUC(X)		((_uc *)(X))
+#define _PUS(X)		((_us *)(X))
+#define _PSL(X)		((_sl *)(X))
+#define _PUL(X)		((_ul *)(X))
 
 #define _B(B)		x86_emit_byte((B))
 #define _W(W)		x86_emit_word((W))
@@ -399,15 +399,15 @@ typedef unsigned int	_ul;
 /* --- Memory subformats - urgh! ------------------------------------------- */
 
 /* _r_D() is RIP addressing mode if X86_TARGET_64BIT, use _r_DSIB() instead */
-#define _r_D(	R, D	  )	(_Mrm(_b00,_rN(R),_b101 )		             ,_L((long)(D)))
-#define _r_DSIB(R, D      )	(_Mrm(_b00,_rN(R),_b100 ),_SIB(_SCL(1),_b100 ,_b101 ),_L((long)(D)))
+#define _r_D(	R, D	  )	(_Mrm(_b00,_rN(R),_b101 )		             ,_L((_sl)(D)))
+#define _r_DSIB(R, D      )	(_Mrm(_b00,_rN(R),_b100 ),_SIB(_SCL(1),_b100 ,_b101 ),_L((_sl)(D)))
 #define _r_0B(	R,   B    )	(_Mrm(_b00,_rN(R),_rA(B))			                   )
 #define _r_0BIS(R,   B,I,S)	(_Mrm(_b00,_rN(R),_b100 ),_SIB(_SCL(S),_rA(I),_rA(B))              )
-#define _r_1B(	R, D,B    )	(_Mrm(_b01,_rN(R),_rA(B))		             ,_B((long)(D)))
-#define _r_1BIS(R, D,B,I,S)	(_Mrm(_b01,_rN(R),_b100 ),_SIB(_SCL(S),_rA(I),_rA(B)),_B((long)(D)))
-#define _r_4B(	R, D,B    )	(_Mrm(_b10,_rN(R),_rA(B))		             ,_L((long)(D)))
-#define _r_4IS( R, D,I,S)	(_Mrm(_b00,_rN(R),_b100 ),_SIB(_SCL(S),_rA(I),_b101 ),_L((long)(D)))
-#define _r_4BIS(R, D,B,I,S)	(_Mrm(_b10,_rN(R),_b100 ),_SIB(_SCL(S),_rA(I),_rA(B)),_L((long)(D)))
+#define _r_1B(	R, D,B    )	(_Mrm(_b01,_rN(R),_rA(B))		             ,_B((_sc)(D)))
+#define _r_1BIS(R, D,B,I,S)	(_Mrm(_b01,_rN(R),_b100 ),_SIB(_SCL(S),_rA(I),_rA(B)),_B((_sc)(D)))
+#define _r_4B(	R, D,B    )	(_Mrm(_b10,_rN(R),_rA(B))		             ,_L((_sl)(D)))
+#define _r_4IS( R, D,I,S)	(_Mrm(_b00,_rN(R),_b100 ),_SIB(_SCL(S),_rA(I),_b101 ),_L((_sl)(D)))
+#define _r_4BIS(R, D,B,I,S)	(_Mrm(_b10,_rN(R),_b100 ),_SIB(_SCL(S),_rA(I),_rA(B)),_L((_sl)(D)))
 
 #define _r_DB(  R, D,B    )	((_s0P(D) && (!_rbp13P(B)) ? _r_0B  (R,  B    ) : (_s8P(D) ? _r_1B(  R,D,B    ) : _r_4B(  R,D,B    ))))
 #define _r_DBIS(R, D,B,I,S)	((_s0P(D) && (!_rbp13P(B)) ? _r_0BIS(R,  B,I,S) : (_s8P(D) ? _r_1BIS(R,D,B,I,S) : _r_4BIS(R,D,B,I,S))))
@@ -439,8 +439,8 @@ typedef unsigned int	_ul;
 #define	 _d16()					   (		  _B(0x66	)				  )
 #define	  _O(	     OP				)  (		  _B(  OP	)				  )
 #define	  _Or(	     OP,R			)  (		  _B( (OP)|_r(R))				  )
-#define	 _OO(	     OP				)  ( _B((OP)>>8), _B( (OP)	)				  )
-#define	 _OOr(	     OP,R			)  ( _B((OP)>>8), _B( (OP)|_r(R))				  )
+#define	 _OO(	     OP				)  ( _B((OP)>>8), _B(( (OP)      )&0xff)			  )
+#define	 _OOr(	     OP,R			)  ( _B((OP)>>8), _B(( (OP)|_r(R))&0xff)			  )
 #define	  _Os(	     OP,B			)  (	_s8P(B) ? _B(((OP)|_b10)) : _B(OP)			  )
 #define	    _sW(			     W	)  (				       _s8P(W) ? _B(W):_W(W)	  )
 #define	    _sL(			     L	)  (				       _s8P(L) ? _B(L):_L(L)	  )
@@ -449,6 +449,7 @@ typedef unsigned int	_ul;
 #define	  _O_B(	     OP			    ,B	)  (	    _O	    (  OP  )			      ,_B(B)	  )
 #define	  _O_W(	     OP			    ,W	)  (	    _O	    (  OP  )			      ,_W(W)	  )
 #define	  _O_L(	     OP			    ,L	)  (	    _O	    (  OP  )			      ,_L(L)	  )
+#define	  _OO_L(     OP			    ,L	)  (	   _OO	    (  OP  )			      ,_L(L)	  )
 #define	  _O_D8(     OP			    ,D	)  (	    _O	    (  OP  )			     ,_D8(D)	  )
 #define	  _O_D32(     OP		    ,D	)  (	    _O	    (  OP  )			     ,_D32(D)	  )
 #define	 _OO_D32(     OP		    ,D	)  (	   _OO	    (  OP  )			     ,_D32(D)	  )
@@ -543,8 +544,8 @@ enum {
 /*									_format		Opcd		,Mod ,r	    ,m		,mem=dsp+sib	,imm... */
 
 #define _ALUBrr(OP,RS, RD)		(_REXBrr(RS, RD),		_O_Mrm		(((OP) << 3)	,_b11,_r1(RS),_r1(RD)				))
-#define _ALUBmr(OP, MD, MB, MI, MS, RD)	(_REXBmr(MB, MI, RD),		_O_r_X		(((OP) << 3) + 2,_r1(RD)		,MD,MB,MI,MS		))
-#define _ALUBrm(OP, RS, MD, MB, MI, MS)	(_REXBrm(RS, MB, MI),		_O_r_X		(((OP) << 3)	,    ,_r1(RS)		,MD,MB,MI,MS		))
+#define _ALUBmr(OP, MD, MB, MI, MS, RD)	(_REXBmr(MB, MI, RD),		_O_r_X		(((OP) << 3) + 2     ,_r1(RD)		,MD,MB,MI,MS		))
+#define _ALUBrm(OP, RS, MD, MB, MI, MS)	(_REXBrm(RS, MB, MI),		_O_r_X		(((OP) << 3)	     ,_r1(RS)		,MD,MB,MI,MS		))
 #define _ALUBir(OP, IM, RD)		(X86_OPTIMIZE_ALU && ((RD) == X86_AL) ? \
 					(_REXBrr(0, RD),		_O_B		(((OP) << 3) + 4					,_su8(IM))) : \
 					(_REXBrr(0, RD),		_O_Mrm_B	(0x80		,_b11,OP     ,_r1(RD)			,_su8(IM))) )
@@ -1021,7 +1022,7 @@ enum {
 #define _BTQrm(OP, RS, MD, MB, MI, MS)	(_REXQrm(RS, MB, MI),		_OO_r_X		(0x0f83|((OP)<<3)     ,_r8(RS)		,MD,MB,MI,MS		))
 
 #define BTWir(IM, RD)			_BTWir(X86_BT, IM, RD)
-#define BTWim(IM, MD, MB, MI, MS)	_BTWim(X86_BT, IM, MD, MI, MS)
+#define BTWim(IM, MD, MB, MI, MS)	_BTWim(X86_BT, IM, MD, MB, MI, MS)
 #define BTWrr(RS, RD)			_BTWrr(X86_BT, RS, RD)
 #define BTWrm(RS, MD, MB, MI, MS)	_BTWrm(X86_BT, RS, MD, MB, MI, MS)
 
@@ -1036,7 +1037,7 @@ enum {
 #define BTQrm(RS, MD, MB, MI, MS)	_BTQrm(X86_BT, RS, MD, MB, MI, MS)
 
 #define BTCWir(IM, RD)			_BTWir(X86_BTC, IM, RD)
-#define BTCWim(IM, MD, MB, MI, MS)	_BTWim(X86_BTC, IM, MD, MI, MS)
+#define BTCWim(IM, MD, MB, MI, MS)	_BTWim(X86_BTC, IM, MD, MB, MI, MS)
 #define BTCWrr(RS, RD)			_BTWrr(X86_BTC, RS, RD)
 #define BTCWrm(RS, MD, MB, MI, MS)	_BTWrm(X86_BTC, RS, MD, MB, MI, MS)
 
@@ -1051,7 +1052,7 @@ enum {
 #define BTCQrm(RS, MD, MB, MI, MS)	_BTQrm(X86_BTC, RS, MD, MB, MI, MS)
 
 #define BTRWir(IM, RD)			_BTWir(X86_BTR, IM, RD)
-#define BTRWim(IM, MD, MB, MI, MS)	_BTWim(X86_BTR, IM, MD, MI, MS)
+#define BTRWim(IM, MD, MB, MI, MS)	_BTWim(X86_BTR, IM, MD, MB, MI, MS)
 #define BTRWrr(RS, RD)			_BTWrr(X86_BTR, RS, RD)
 #define BTRWrm(RS, MD, MB, MI, MS)	_BTWrm(X86_BTR, RS, MD, MB, MI, MS)
 
@@ -1066,7 +1067,7 @@ enum {
 #define BTRQrm(RS, MD, MB, MI, MS)	_BTQrm(X86_BTR, RS, MD, MB, MI, MS)
 
 #define BTSWir(IM, RD)			_BTWir(X86_BTS, IM, RD)
-#define BTSWim(IM, MD, MB, MI, MS)	_BTWim(X86_BTS, IM, MD, MI, MS)
+#define BTSWim(IM, MD, MB, MI, MS)	_BTWim(X86_BTS, IM, MD, MB, MI, MS)
 #define BTSWrr(RS, RD)			_BTWrr(X86_BTS, RS, RD)
 #define BTSWrm(RS, MD, MB, MI, MS)	_BTWrm(X86_BTS, RS, MD, MB, MI, MS)
 
@@ -1579,10 +1580,10 @@ enum {
 #define MOVZWLrr(RS, RD)		(_REXLrr(RD, RS),		_OO_Mrm		(0x0fb7		,_b11,_r4(RD),_r2(RS)				))
 #define MOVZWLmr(MD, MB, MI, MS, RD)	(_REXLmr(MB, MI, RD),		_OO_r_X		(0x0fb7		     ,_r4(RD)		,MD,MB,MI,MS		))
 
-#define MOVSWQrr(RS, RD)		(_REXQrr(RD, RS),		_OO_Mrm		(0x0fbf		,_b11,_r8(RD),_r2(RS)				))
-#define MOVSWQmr(MD, MB, MI, MS, RD)	(_REXQmr(MB, MI, RD),		_OO_r_X		(0x0fbf		     ,_r8(RD)		,MD,MB,MI,MS		))
-#define MOVZWQrr(RS, RD)		(_REXQrr(RD, RS),		_OO_Mrm		(0x0fb7		,_b11,_r8(RD),_r2(RS)				))
-#define MOVZWQmr(MD, MB, MI, MS, RD)	(_REXQmr(MB, MI, RD),		_OO_r_X		(0x0fb7		     ,_r8(RD)		,MD,MB,MI,MS		))
+#define MOVSWQrr(RS, RD)		_m64only((_REXQrr(RD, RS),	_OO_Mrm		(0x0fbf		,_b11,_r8(RD),_r2(RS)				)))
+#define MOVSWQmr(MD, MB, MI, MS, RD)	_m64only((_REXQmr(MB, MI, RD),	_OO_r_X		(0x0fbf		     ,_r8(RD)		,MD,MB,MI,MS		)))
+#define MOVZWQrr(RS, RD)		_m64only((_REXQrr(RD, RS),	_OO_Mrm		(0x0fb7		,_b11,_r8(RD),_r2(RS)				)))
+#define MOVZWQmr(MD, MB, MI, MS, RD)	_m64only((_REXQmr(MB, MI, RD),	_OO_r_X		(0x0fb7		     ,_r8(RD)		,MD,MB,MI,MS		)))
 
 #define MOVSLQrr(RS, RD)		_m64only((_REXQrr(RD, RS),	_O_Mrm		(0x63		,_b11,_r8(RD),_r4(RS)				)))
 #define MOVSLQmr(MD, MB, MI, MS, RD)	_m64only((_REXQmr(MB, MI, RD),	_O_r_X		(0x63		     ,_r8(RD)		,MD,MB,MI,MS		)))
@@ -1590,14 +1591,15 @@ enum {
 /*									_format		Opcd		,Mod ,r	    ,m		,mem=dsp+sib	,imm... */
 
 #define LEALmr(MD, MB, MI, MS, RD)	(_REXLmr(MB, MI, RD),		_O_r_X		(0x8d		     ,_r4(RD)		,MD,MB,MI,MS		))
+#define LEAQmr(MD, MB, MI, MS, RD)	(_REXQmr(MB, MI, RD),		_O_r_X		(0x8d		     ,_r4(RD)		,MD,MB,MI,MS		))
 
 #define BSWAPLr(R)			(_REXLrr(0, R),			_OOr		(0x0fc8,_r4(R)							))
 #define BSWAPQr(R)			(_REXQrr(0, R),			_OOr		(0x0fc8,_r8(R)							))
 
 #define CLC()								_O		(0xf8								)
 #define STC()								_O		(0xf9								)
-
 #define CMC()								_O		(0xf5								)
+
 #define CLD()								_O		(0xfc								)
 #define STD()								_O		(0xfd								)
 
@@ -1617,8 +1619,8 @@ enum {
 #define CDQ				CLTD
 #define CQO				CQTO
 
-#define LAHF()				_m32only(			_O		(0x9f								))
-#define SAHF()				_m32only(			_O		(0x9e								))
+#define LAHF()								_O		(0x9f								)
+#define SAHF()								_O		(0x9e								)
 
 /*									_format		Opcd		,Mod ,r	    ,m		,mem=dsp+sib	,imm... */
 
@@ -1634,13 +1636,34 @@ enum {
 #define NOP()								_O		(0x90								)
 
 
+/* --- Media 64-bit instructions ------------------------------------------- */
+
+#define EMMS()								_OO		(0x0f77								)
+
+
 /* --- Media 128-bit instructions ------------------------------------------ */
 
+enum {
+  X86_SSE_CC_EQ		= 0,
+  X86_SSE_CC_LT		= 1,
+  X86_SSE_CC_GT		= 1,
+  X86_SSE_CC_LE		= 2,
+  X86_SSE_CC_GE		= 2,
+  X86_SSE_CC_U		= 3,
+  X86_SSE_CC_NE		= 4,
+  X86_SSE_CC_NLT	= 5,
+  X86_SSE_CC_NGT	= 5,
+  X86_SSE_CC_NLE	= 6,
+  X86_SSE_CC_NGE	= 6,
+  X86_SSE_CC_O		= 7
+};
+
 enum {
   X86_SSE_CVTIS  = 0x2a,
   X86_SSE_CVTSI  = 0x2d,
   X86_SSE_UCOMI  = 0x2e,
   X86_SSE_COMI   = 0x2f,
+  X86_SSE_CMP		= 0xc2,
   X86_SSE_SQRT   = 0x51,
   X86_SSE_RSQRT  = 0x52,
   X86_SSE_RCP    = 0x53,
@@ -1656,13 +1679,76 @@ enum {
   X86_SSE_MIN    = 0x5d,
   X86_SSE_DIV    = 0x5e,
   X86_SSE_MAX    = 0x5f,
+  X86_SSE_MOVMSK	= 0x50,
+  X86_SSE_PACKSSDW	= 0x6b,
+  X86_SSE_PACKSSWB	= 0x63,
+  X86_SSE_PACKUSWB	= 0x67,
+  X86_SSE_PADDB		= 0xfc,
+  X86_SSE_PADDD		= 0xfe,
+  X86_SSE_PADDQ		= 0xd4,
+  X86_SSE_PADDSB	= 0xec,
+  X86_SSE_PADDSW	= 0xed,
+  X86_SSE_PADDUSB	= 0xdc,
+  X86_SSE_PADDUSW	= 0xdd,
+  X86_SSE_PADDW		= 0xfd,
+  X86_SSE_PAND		= 0xdb,
+  X86_SSE_PANDN		= 0xdf,
+  X86_SSE_PAVGB		= 0xe0,
+  X86_SSE_PAVGW		= 0xe3,
+  X86_SSE_PCMPEQB	= 0x74,
+  X86_SSE_PCMPEQD	= 0x76,
+  X86_SSE_PCMPEQW	= 0x75,
+  X86_SSE_PCMPGTB	= 0x64,
+  X86_SSE_PCMPGTD	= 0x66,
+  X86_SSE_PCMPGTW	= 0x65,
+  X86_SSE_PMADDWD	= 0xf5,
+  X86_SSE_PMAXSW	= 0xee,
+  X86_SSE_PMAXUB	= 0xde,
+  X86_SSE_PMINSW	= 0xea,
+  X86_SSE_PMINUB	= 0xda,
+  X86_SSE_PMOVMSKB	= 0xd7,
+  X86_SSE_PMULHUW	= 0xe4,
+  X86_SSE_PMULHW	= 0xe5,
+  X86_SSE_PMULLW	= 0xd5,
+  X86_SSE_PMULUDQ	= 0xf4,
+  X86_SSE_POR		= 0xeb,
+  X86_SSE_PSADBW	= 0xf6,
+  X86_SSE_PSLLD		= 0xf2,
+  X86_SSE_PSLLQ		= 0xf3,
+  X86_SSE_PSLLW		= 0xf1,
+  X86_SSE_PSRAD		= 0xe2,
+  X86_SSE_PSRAW		= 0xe1,
+  X86_SSE_PSRLD		= 0xd2,
+  X86_SSE_PSRLQ		= 0xd3,
+  X86_SSE_PSRLW		= 0xd1,
+  X86_SSE_PSUBB		= 0xf8,
+  X86_SSE_PSUBD		= 0xfa,
+  X86_SSE_PSUBQ		= 0xfb,
+  X86_SSE_PSUBSB	= 0xe8,
+  X86_SSE_PSUBSW	= 0xe9,
+  X86_SSE_PSUBUSB	= 0xd8,
+  X86_SSE_PSUBUSW	= 0xd9,
+  X86_SSE_PSUBW		= 0xf9,
+  X86_SSE_PUNPCKHBW	= 0x68,
+  X86_SSE_PUNPCKHDQ	= 0x6a,
+  X86_SSE_PUNPCKHQDQ	= 0x6d,
+  X86_SSE_PUNPCKHWD	= 0x69,
+  X86_SSE_PUNPCKLBW	= 0x60,
+  X86_SSE_PUNPCKLDQ	= 0x62,
+  X86_SSE_PUNPCKLQDQ	= 0x6c,
+  X86_SSE_PUNPCKLWD	= 0x61,
+  X86_SSE_PXOR		= 0xef,
 };
 
 /*									_format		Opcd		,Mod ,r	     ,m		,mem=dsp+sib	,imm... */
 
+#define __SSELir(OP,MO,IM,RD)		(_REXLrr(0, RD),		_OO_Mrm_B	(0x0f00|(OP)	,_b11,MO     ,_rX(RD)			,_u8(IM)))
+#define __SSELim(OP,MO,IM,MD,MB,MI,MS)	(_REXLrm(0, MB, MI),		_OO_r_X_B	(0x0f00|(OP)	     ,MO		,MD,MB,MI,MS	,_u8(IM)))
 #define __SSELrr(OP,RS,RSA,RD,RDA)	(_REXLrr(RD, RS),		_OO_Mrm		(0x0f00|(OP)	,_b11,RDA(RD),RSA(RS)				))
 #define __SSELmr(OP,MD,MB,MI,MS,RD,RDA)	(_REXLmr(MB, MI, RD),		_OO_r_X		(0x0f00|(OP)	     ,RDA(RD)		,MD,MB,MI,MS		))
 #define __SSELrm(OP,RS,RSA,MD,MB,MI,MS)	(_REXLrm(RS, MB, MI),		_OO_r_X		(0x0f00|(OP)	     ,RSA(RS)		,MD,MB,MI,MS		))
+#define __SSELirr(OP,IM,RS,RD)		(_REXLrr(RD, RS),		_OO_Mrm_B	(0x0f00|(OP)	,_b11,_rX(RD),_rX(RS)			,_u8(IM)))
+#define __SSELimr(OP,IM,MD,MB,MI,MS,RD)	(_REXLmr(MB, MI, RD),		_OO_r_X_B	(0x0f00|(OP)	     ,_rX(RD)		,MD,MB,MI,MS	,_u8(IM)))
 
 #define __SSEQrr(OP,RS,RSA,RD,RDA)	(_REXQrr(RD, RS),		_OO_Mrm		(0x0f00|(OP)	,_b11,RDA(RD),RSA(RS)				))
 #define __SSEQmr(OP,MD,MB,MI,MS,RD,RDA)	(_REXQmr(MB, MI, RD),		_OO_r_X		(0x0f00|(OP)	     ,RDA(RD)		,MD,MB,MI,MS		))
@@ -1671,6 +1757,10 @@ enum {
 #define _SSELrr(PX,OP,RS,RSA,RD,RDA)					(_B(PX), __SSELrr(OP, RS, RSA, RD, RDA))
 #define _SSELmr(PX,OP,MD,MB,MI,MS,RD,RDA)				(_B(PX), __SSELmr(OP, MD, MB, MI, MS, RD, RDA))
 #define _SSELrm(PX,OP,RS,RSA,MD,MB,MI,MS)				(_B(PX), __SSELrm(OP, RS, RSA, MD, MB, MI, MS))
+#define _SSELir(PX,OP,MO,IM,RD)						(_B(PX), __SSELir(OP, MO, IM, RD))
+#define _SSELim(PX,OP,MO,IM,MD,MB,MI,MS)				(_B(PX), __SSELim(OP, MO, IM, MD, MB, MI, MS))
+#define _SSELirr(PX,OP,IM,RS,RD)					(_B(PX), __SSELirr(OP, IM, RS, RD))
+#define _SSELimr(PX,OP,IM,MD,MB,MI,MS,RD)				(_B(PX), __SSELimr(OP, IM, MD, MB, MI, MS, RD))
 
 #define _SSEQrr(PX,OP,RS,RSA,RD,RDA)					(_B(PX), __SSEQrr(OP, RS, RSA, RD, RDA))
 #define _SSEQmr(PX,OP,MD,MB,MI,MS,RD,RDA)				(_B(PX), __SSEQmr(OP, MD, MB, MI, MS, RD, RDA))
@@ -1679,18 +1769,26 @@ enum {
 #define _SSEPSrr(OP,RS,RD)		__SSELrr(      OP, RS,_rX, RD,_rX)
 #define _SSEPSmr(OP,MD,MB,MI,MS,RD)	__SSELmr(      OP, MD, MB, MI, MS, RD,_rX)
 #define _SSEPSrm(OP,RS,MD,MB,MI,MS)	__SSELrm(      OP, RS,_rX, MD, MB, MI, MS)
+#define _SSEPSirr(OP,IM,RS,RD)		__SSELirr(     OP, IM, RS, RD)
+#define _SSEPSimr(OP,IM,MD,MB,MI,MS,RD)	__SSELimr(     OP, IM, MD, MB, MI, MS, RD)
 
 #define _SSEPDrr(OP,RS,RD)		 _SSELrr(0x66, OP, RS,_rX, RD,_rX)
 #define _SSEPDmr(OP,MD,MB,MI,MS,RD)	 _SSELmr(0x66, OP, MD, MB, MI, MS, RD,_rX)
 #define _SSEPDrm(OP,RS,MD,MB,MI,MS)	 _SSELrm(0x66, OP, RS,_rX, MD, MB, MI, MS)
+#define _SSEPDirr(OP,IM,RS,RD)		 _SSELirr(0x66, OP, IM, RS, RD)
+#define _SSEPDimr(OP,IM,MD,MB,MI,MS,RD)	 _SSELimr(0x66, OP, IM, MD, MB, MI, MS, RD)
 
 #define _SSESSrr(OP,RS,RD)		 _SSELrr(0xf3, OP, RS,_rX, RD,_rX)
 #define _SSESSmr(OP,MD,MB,MI,MS,RD)	 _SSELmr(0xf3, OP, MD, MB, MI, MS, RD,_rX)
 #define _SSESSrm(OP,RS,MD,MB,MI,MS)	 _SSELrm(0xf3, OP, RS,_rX, MD, MB, MI, MS)
+#define _SSESSirr(OP,IM,RS,RD)		 _SSELirr(0xf3, OP, IM, RS, RD)
+#define _SSESSimr(OP,IM,MD,MB,MI,MS,RD)	 _SSELimr(0xf3, OP, IM, MD, MB, MI, MS, RD)
 
 #define _SSESDrr(OP,RS,RD)		 _SSELrr(0xf2, OP, RS,_rX, RD,_rX)
 #define _SSESDmr(OP,MD,MB,MI,MS,RD)	 _SSELmr(0xf2, OP, MD, MB, MI, MS, RD,_rX)
 #define _SSESDrm(OP,RS,MD,MB,MI,MS)	 _SSELrm(0xf2, OP, RS,_rX, MD, MB, MI, MS)
+#define _SSESDirr(OP,IM,RS,RD)		 _SSELirr(0xf2, IM, OP, RS, RD)
+#define _SSESDimr(OP,IM,MD,MB,MI,MS,RD)	 _SSELimr(0xf2, IM, OP, MD, MB, MI, MS, RD)
 
 #define ADDPSrr(RS, RD)			_SSEPSrr(X86_SSE_ADD, RS, RD)
 #define ADDPSmr(MD, MB, MI, MS, RD)	_SSEPSmr(X86_SSE_ADD, MD, MB, MI, MS, RD)
@@ -1712,6 +1810,16 @@ enum {
 #define ANDPDrr(RS, RD)			_SSEPDrr(X86_SSE_AND, RS, RD)
 #define ANDPDmr(MD, MB, MI, MS, RD)	_SSEPDmr(X86_SSE_AND, MD, MB, MI, MS, RD)
 
+#define CMPPSrr(IM, RS, RD)		_SSEPSirr(X86_SSE_CMP, IM, RS, RD)
+#define CMPPSmr(IM, MD, MB, MI, MS, RD)	_SSEPSimr(X86_SSE_CMP, IM, MD, MB, MI, MS, RD)
+#define CMPPDrr(IM, RS, RD)		_SSEPDirr(X86_SSE_CMP, IM, RS, RD)
+#define CMPPDmr(IM, MD, MB, MI, MS, RD)	_SSEPDimr(X86_SSE_CMP, IM, MD, MB, MI, MS, RD)
+
+#define CMPSSrr(IM, RS, RD)		_SSESSirr(X86_SSE_CMP, IM, RS, RD)
+#define CMPSSmr(IM, MD, MB, MI, MS, RD)	_SSESSimr(X86_SSE_CMP, IM, MD, MB, MI, MS, RD)
+#define CMPSDrr(IM, RS, RD)		_SSESDirr(X86_SSE_CMP, IM, RS, RD)
+#define CMPSDmr(IM, MD, MB, MI, MS, RD)	_SSESDimr(X86_SSE_CMP, IM, MD, MB, MI, MS, RD)
+
 #define DIVPSrr(RS, RD)			_SSEPSrr(X86_SSE_DIV, RS, RD)
 #define DIVPSmr(MD, MB, MI, MS, RD)	_SSEPSmr(X86_SSE_DIV, MD, MB, MI, MS, RD)
 #define DIVPDrr(RS, RD)			_SSEPDrr(X86_SSE_DIV, RS, RD)
@@ -1871,6 +1979,9 @@ enum {
 #define MOVDMQrm(RS, MD, MB, MI, MS)	__SSEQrm(      0x7e, RS,_rM, MD, MB, MI, MS)
 
 #define MOVDQ2Qrr(RS, RD)		 _SSELrr(0xf2, 0xd6, RS,_rX, RD,_rM)
+#define MOVMSKPSrr(RS, RD)		__SSELrr(      0x50, RS,_rX, RD,_r4)
+#define MOVMSKPDrr(RS, RD)		 _SSELrr(0x66, 0x50, RS,_rX, RD,_r4)
+
 #define MOVHLPSrr(RS, RD)		__SSELrr(      0x12, RS,_rX, RD,_rX)
 #define MOVLHPSrr(RS, RD)		__SSELrr(      0x16, RS,_rX, RD,_rX)
 
